#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

import warnings

from dataclasses import dataclass
from io import BufferedWriter
from pathlib import Path
from typing import Dict, Optional
import os
from urllib.parse import quote

from requests import Response
from requests.structures import CaseInsensitiveDict
from aistore.sdk.archive_config import ArchiveConfig
from aistore.sdk.blob_download_config import BlobDownloadConfig
from aistore.sdk.etl import ETLConfig
from aistore.sdk.const import (
    BYTE_RANGE_PREFIX_LENGTH,
    DEFAULT_CHUNK_SIZE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_HEAD,
    QPARAM_ARCHPATH,
    QPARAM_ARCHREGX,
    QPARAM_ARCHMODE,
    QPARAM_ETL_NAME,
    QPARAM_ETL_ARGS,
    QPARAM_LATEST,
    QPARAM_PROPS,
    QPARAM_SYNC,
    QPARAM_OBJ_TO,
    ACT_PROMOTE,
    HTTP_METHOD_POST,
    HTTP_METHOD_PUT,
    URL_PATH_OBJECTS,
    HEADER_RANGE,
    ACT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_WORKERS,
    HEADER_OBJECT_BLOB_CHUNK_SIZE,
)
from aistore.sdk.provider import Provider
from aistore.sdk.obj.object_client import ObjectClient
from aistore.sdk.obj.object_reader import ObjectReader
from aistore.sdk.obj.object_writer import ObjectWriter
from aistore.sdk.obj.multipart_upload import MultipartUpload
from aistore.sdk.request_client import RequestClient
from aistore.sdk.types import (
    ActionMsg,
    PromoteAPIArgs,
    BlobMsg,
)
from aistore.sdk.obj.object_props import ObjectProps
from aistore.sdk.obj.object_attributes import ObjectAttributesV2


@dataclass
class BucketDetails:
    """
    Metadata about a bucket, used by objects within that bucket.
    """

    name: str
    provider: Provider
    qparams: Dict[str, str]
    path: str


# pylint: disable=too-many-public-methods
class Object:
    """
    Provides methods for interacting with an object in AIS.

    Args:
        client (RequestClient): Client used for all http requests.
        bck_details (BucketDetails): Metadata about the bucket to which this object belongs.
        name (str): Name of the object.
        props (ObjectProps, optional): Properties of the object, as updated by head(), optionally pre-initialized.
    """

    def __init__(
        self,
        client: RequestClient,
        bck_details: BucketDetails,
        name: str,
        props: Optional[ObjectProps] = None,
    ):
        self._client = client
        self._bck_details = bck_details
        self._bck_path = f"{URL_PATH_OBJECTS}/{ bck_details.name}"
        self._name = name
        self._props = props
        self._object_path = f"{self._bck_path}/{quote(name)}"

    @property
    def bucket_name(self) -> str:
        """Name of the bucket where this object resides."""
        return self._bck_details.name

    @property
    def bucket_provider(self) -> Provider:
        """Provider of the bucket where this object resides (e.g. ais, s3, gcp)."""
        return self._bck_details.provider

    @property
    def query_params(self) -> Dict[str, str]:
        """Query params used as a base for constructing all requests for this object."""
        return self._bck_details.qparams

    @property
    def name(self) -> str:
        """Name of this object."""
        return self._name

    @property
    def uname(self) -> str:
        """
        Unified name (uname) of this object, which combines the bucket path and object name.

        Returns:
            str: The unified name in the format bucket_path/object_name
        """
        return os.path.join(self._bck_details.path, self.name)

    @property
    def props(self) -> ObjectProps:
        """
        Get the latest properties of the object.

        This will make a HEAD request to the AIStore cluster to fetch up-to-date object headers
        and refresh the internal `_props` cache. Use this when you want to ensure you're accessing
        the most recent metadata for the object.

        Returns:
            ObjectProps: The latest object properties from the server.
        """
        self.head()
        # Head must always set _props
        assert self._props is not None
        return self._props

    @property
    def props_cached(self) -> Optional[ObjectProps]:
        """
        Get the cached object properties (without making a network call).

        This is useful when:
        - You want to avoid a network request.
        - You're sure the cached `_props` was already set via a previous call to `head()` or during object construction.

        Returns:
            ObjectProps or None: Cached object properties, or None if not set.
        """
        return self._props

    def head(self) -> CaseInsensitiveDict:
        """
        Requests object properties and returns headers. Updates props.

        Deprecation notice:
            This is the legacy HEAD(object) v1 call. It remains supported in AIS 4.2,
            but new development should target Object HEAD v2 when available in the SDK.
            The v1 path is planned for removal in a future major release.

        Returns:
            Response header with the object properties.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        headers = self._client.request(
            HTTP_METHOD_HEAD,
            path=self._object_path,
            params=self.query_params,
        ).headers
        self._props = ObjectProps(headers)
        return headers

    def head_v2(self, props: str = "") -> ObjectAttributesV2:
        """
        Make a HEAD request with selective property retrieval (V2 API).

        EXPERIMENTAL: This API is experimental and may change in future releases.

        This method allows requesting specific object properties, reducing
        response size and processing overhead when only certain attributes
        are needed.

        Args:
            props: Comma-separated list of properties to retrieve.
                   Available values: name, size, version, checksum, atime, present,
                   copies, ec, custom, location, chunked, last-modified, etag.
                   See: https://github.com/NVIDIA/aistore/blob/main/api/apc/lsmsg.go
                   If empty, returns default properties (name, size).


        Returns:
            ObjectAttributesV2: Parsed V2 object attributes (includes chunk info, last-modified, etag).
        """
        params = self.query_params.copy()
        # Always set props to trigger V2 endpoint; default to "name,size"
        params[QPARAM_PROPS] = props if props else "name,size"

        headers = self._client.request(
            HTTP_METHOD_HEAD,
            path=self._object_path,
            params=params,
        ).headers
        return ObjectAttributesV2(headers)

    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    def get_reader(
        self,
        archive_config: Optional[ArchiveConfig] = None,
        blob_download_config: Optional[BlobDownloadConfig] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: Optional[ETLConfig] = None,
        writer: Optional[BufferedWriter] = None,
        latest: bool = False,
        byte_range: Optional[str] = None,
        direct: bool = False,
        num_workers: Optional[int] = None,
    ) -> ObjectReader:
        """
        Creates and returns an ObjectReader with access to object contents
        and optionally writes to a provided writer.

        Args:
            archive_config (Optional[ArchiveConfig]): Settings for archive extraction.
            blob_download_config (Optional[BlobDownloadConfig]): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (Optional[ETLConfig]): Settings for ETL-specific operations (name, args).
            writer (Optional[BufferedWriter]): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (Optional[str]): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.
            direct (bool, optional): If True, the object content is read directly from the target node,
                bypassing the proxy.
            num_workers (Optional[int]): If provided, use concurrent range-reads with this many
                workers for faster downloads.

        Returns:
            ObjectReader: An iterator for streaming object content.

        Raises:
            ValueError: If `byte_range` is used with `blob_download_config`.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """

        params = self.query_params.copy()
        headers = {}
        byte_range_tuple = (None, None)

        # Archive Configuration
        if archive_config:
            if archive_config.mode:
                params[QPARAM_ARCHMODE] = archive_config.mode.value
            params.update(
                {
                    QPARAM_ARCHPATH: archive_config.archpath,
                    QPARAM_ARCHREGX: archive_config.regex,
                }
            )

        # Blob Download Configuration
        if blob_download_config:
            headers.update(
                {
                    HEADER_OBJECT_BLOB_DOWNLOAD: "true",
                    HEADER_OBJECT_BLOB_CHUNK_SIZE: blob_download_config.chunk_size,
                    HEADER_OBJECT_BLOB_WORKERS: blob_download_config.num_workers,
                }
            )

        # ETL Configuration
        if etl:
            etl.update_qparams(params)

        # Latest Object Version
        if latest:
            params[QPARAM_LATEST] = "true"

        # Byte Range Validation
        if byte_range and blob_download_config:
            raise ValueError("Cannot use `byte_range` with `blob_download_config`.")

        # Parallel download validation
        if num_workers is not None:
            if num_workers < 2:
                raise ValueError("`num_workers` must be at least 2.")
            if byte_range:
                raise ValueError(
                    "Cannot use `num_workers` with `byte_range`. "
                    "Parallel download is for full object retrieval."
                )
            if blob_download_config:
                raise ValueError(
                    "Cannot use `num_workers` with `blob_download_config`."
                )

        if byte_range:
            # For range formatting, see the spec:
            # https://www.rfc-editor.org/rfc/rfc7233#section-2.1
            headers = {HEADER_RANGE: byte_range}
            # Extract left (range_l) and right (range_r) bounds from the byte range string
            headers[HEADER_RANGE] = byte_range
            byte_range_l, _, byte_range_r = byte_range[
                BYTE_RANGE_PREFIX_LENGTH:
            ].partition("-")
            byte_range_tuple = (
                int(byte_range_l) if byte_range_l else None,
                int(byte_range_r) if byte_range_r else None,
            )

        # Object Client
        obj_client = ObjectClient(
            request_client=self._client,
            path=self._object_path,
            params=params,
            headers=headers,
            byte_range=byte_range_tuple,
            uname=self.uname if direct else None,
        )

        obj_reader = ObjectReader(
            object_client=obj_client,
            chunk_size=chunk_size,
            num_workers=num_workers,
        )

        if writer:
            writer.writelines(obj_reader)

        return obj_reader

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def get(
        self,
        archive_config: Optional[ArchiveConfig] = None,
        blob_download_config: Optional[BlobDownloadConfig] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: Optional[ETLConfig] = None,
        writer: Optional[BufferedWriter] = None,
        latest: bool = False,
        byte_range: Optional[str] = None,
    ) -> ObjectReader:
        """
        Deprecated: Use 'get_reader' instead.

        Creates and returns an ObjectReader with access to object contents and optionally writes to a provided writer.

        Args:
            archive_config (ArchiveConfig, optional): Settings for archive extraction.
            blob_download_config (BlobDownloadConfig, optional): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).
            writer (BufferedWriter, optional): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (str, optional): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.

        Returns:
            ObjectReader: An ObjectReader that can be iterated over to stream chunks of object content
            or used to read all content directly.

        Raises:
            ValueError: If `byte_range` is used with `blob_download_config`.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """
        warnings.warn(
            "The `Object.get(...)` method is deprecated and will be removed in a future release."
            "Please replace it with `Object.get_reader()` using the new `ObjectReader` API.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_reader(
            archive_config=archive_config,
            blob_download_config=blob_download_config,
            chunk_size=chunk_size,
            etl=etl,
            writer=writer,
            latest=latest,
            byte_range=byte_range,
        )

    def get_semantic_url(self) -> str:
        """
        Get the semantic URL to the object

        Returns:
            Semantic URL to get object
        """

        return f"{self.bucket_provider.value}://{self.bucket_name}/{self._name}"

    def get_url(self, archpath: str = "", etl: Optional[ETLConfig] = None) -> str:
        """
        Get the full url to the object including base url and any query parameters

        Args:
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file
                from the archive
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).

        Returns:
            Full URL to get object

        """
        params = self.query_params.copy()
        if archpath:
            params[QPARAM_ARCHPATH] = archpath

        # ETL Configuration
        if etl:
            params[QPARAM_ETL_NAME] = etl.name
            if etl.args:
                params[QPARAM_ETL_ARGS] = etl.args

        return self._client.get_full_url(self._object_path, params)

    def put_content(self, content: bytes) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_content' instead.

        Puts bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to put as an object.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        warnings.warn(
            "The `Object.put_content(...)` method is deprecated and will be removed in a future release."
            "Please replace it with `Object.get_writer().put_content(...)` using the new `ObjectWriter` API.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_content(content)

    def put_file(self, path: str or Path) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_file' instead.

        Puts a local file as an object to a bucket in AIS storage.

        Args:
            path (str or Path): Path to local file

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: The path provided is not a valid file
        """
        warnings.warn(
            "The `Object.put_file(...)` method is deprecated and will be removed in a future release."
            "Please replace it with `Object.get_writer().put_file(...)` using the new `ObjectWriter` API.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_file(path)

    def get_writer(self) -> ObjectWriter:
        """
        Create an ObjectWriter to write to object contents and attributes.

        Returns:
            An ObjectWriter which can be used to write to an object's contents and attributes.
        """
        return ObjectWriter(self._client, self._object_path, self.query_params)

    def multipart_upload(self) -> MultipartUpload:
        """
        Create a multipart upload for this object.

        Returns:
            MultipartUpload: A multipart upload instance for this object.
        """
        return MultipartUpload(self._client, self._object_path, self.query_params)

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def promote(
        self,
        path: str,
        target_id: str = "",
        recursive: bool = False,
        overwrite_dest: bool = False,
        delete_source: bool = False,
        src_not_file_share: bool = False,
    ) -> str:
        """
        Promotes a file or folder an AIS target can access to a bucket in AIS storage.
        These files can be either on the physical disk of an AIS target itself or on a network file system
        the cluster can access.
        See more info here: https://aiatscale.org/blog/2022/03/17/promote

        Args:
            path (str): Path to file or folder the AIS cluster can reach
            target_id (str, optional): Promote files from a specific target node
            recursive (bool, optional): Recursively promote objects from files in directories inside the path
            overwrite_dest (bool, optional): Overwrite objects already on AIS
            delete_source (bool, optional): Delete the source files when done promoting
            src_not_file_share (bool, optional): Optimize if the source is guaranteed to not be on a file share

        Returns:
            Job ID (as str) that can be used to check the status of the operation, or empty if job is done synchronously

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            AISError: Path does not exist on the AIS cluster storage
        """
        value = PromoteAPIArgs(
            source_path=path,
            object_name=self.name,
            target_id=target_id,
            recursive=recursive,
            overwrite_dest=overwrite_dest,
            delete_source=delete_source,
            src_not_file_share=src_not_file_share,
        ).as_dict()
        json_val = ActionMsg(action=ACT_PROMOTE, name=path, value=value).model_dump()

        return self._client.request(
            HTTP_METHOD_POST,
            path=self._bck_path,
            params=self.query_params,
            json=json_val,
        ).text

    def delete(self) -> Response:
        """
        Delete an object from a bucket.

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        return self._client.request(
            HTTP_METHOD_DELETE,
            path=self._object_path,
            params=self.query_params,
        )

    def copy(
        self,
        to_obj: "Object",
        etl: Optional[ETLConfig] = None,
        latest: bool = False,
        sync: bool = False,
    ) -> Response:
        """
        Copy this object to another object (which specifies the destination bucket and name),
        optionally with ETL transformation.

        Args:
            to_obj (Object): Destination object specifying both the target bucket and object name
            etl (ETLConfig, optional): ETL configuration for transforming the object during copy
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            sync (bool, optional): In addition to the latest, also entails removing remotely deleted objects

        Returns:
            Response: The response from the copy operation

        Raises:
            requests.RequestException: "There's an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError: Service unavailable

        """
        # Create query parameters for the destination
        query_params = self.query_params.copy()

        # Use uname for destination bucket+object
        query_params[QPARAM_OBJ_TO] = to_obj.uname
        query_params[QPARAM_LATEST] = "true" if latest else "false"
        query_params[QPARAM_SYNC] = "true" if sync else "false"

        # Add ETL configuration if provided
        if etl:
            etl.update_qparams(query_params)

        return self._client.request(
            HTTP_METHOD_PUT,
            path=self._object_path,
            params=query_params,
        )

    def blob_download(
        self,
        chunk_size: int = None,
        num_workers: int = None,
        latest: bool = False,
    ) -> str:
        """
        A special facility to download very large remote objects a.k.a. BLOBs
        Returns job ID that for the blob download operation.

        Args:
            chunk_size (int): chunk size in bytes
            num_workers (int): number of concurrent blob-downloading workers (readers)
            latest (bool): GET the latest object version from the associated remote bucket

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
        """
        params = self.query_params.copy()
        value = BlobMsg(
            chunk_size=chunk_size,
            num_workers=num_workers,
            latest=latest,
        ).as_dict()
        json_val = ActionMsg(
            action=ACT_BLOB_DOWNLOAD, value=value, name=self.name
        ).model_dump()
        return self._client.request(
            HTTP_METHOD_POST, path=self._bck_path, params=params, json=json_val
        ).text

    def append_content(
        self, content: bytes, handle: str = "", flush: bool = False
    ) -> str:
        """
        Deprecated: Use 'ObjectWriter.append_content' instead.

        Append bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to append to the object.
            handle (str): Handle string to use for subsequent appends or flush (empty for the first append).
            flush (bool): Whether to flush and finalize the append operation, making the object accessible.

        Returns:
            handle (str): Handle string to pass for subsequent appends or flush.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        warnings.warn(
            "The `Object.append_content(...)` method is deprecated and will be removed in a future release."
            "Please replace it with `Object.get_writer().append_content(...)` using the new `ObjectWriter` API.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().append_content(content, handle, flush)

    def set_custom_props(
        self, custom_metadata: Dict[str, str], replace_existing: bool = False
    ) -> Response:
        """
        Deprecated: Use 'ObjectWriter.set_custom_props' instead.

        Set custom properties for the object.

        Args:
            custom_metadata (Dict[str, str]): Custom metadata key-value pairs.
            replace_existing (bool, optional): Whether to replace existing metadata. Defaults to False.
        """
        warnings.warn(
            "The `Object.set_custom_props(...)` method is deprecated and will be removed in a future release."
            "Please replace it with `Object.get_writer().set_custom_props(...)` using the new `ObjectWriter` API.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().set_custom_props(custom_metadata, replace_existing)



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import os
import asyncio
from urllib.parse import unquote, quote
from typing import Optional, List, Tuple

from fastapi import (
    FastAPI,
    Request,
    HTTPException,
    Response,
    WebSocket,
)
import httpx
import aiofiles
import uvicorn

from aistore.sdk.etl.webserver.base_etl_server import ETLServer
from aistore.sdk.etl.webserver.utils import (
    compose_etl_direct_put_url,
    parse_etl_pipeline,
)
from aistore.sdk.errors import InvalidPipelineError
from aistore.sdk.const import (
    HEADER_NODE_URL,
    HEADER_CONTENT_LENGTH,
    STATUS_OK,
    ETL_WS_FQN,
    ETL_WS_PATH,
    ETL_WS_PIPELINE,
    HEADER_DIRECT_PUT_LENGTH,
    QPARAM_ETL_ARGS,
    QPARAM_ETL_FQN,
    STATUS_INTERNAL_SERVER_ERROR,
)

HTTP_LIMITS = httpx.Limits(
    max_connections=int(os.getenv("MAX_CONN", "256")),
    max_keepalive_connections=int(os.getenv("MAX_KEEPALIVE_CONN", "128")),
    keepalive_expiry=int(os.getenv("KEEPALIVE_EXPIRY", "30")),
)


class FastAPIServer(ETLServer):
    """
    FastAPI server implementation for ETL transformations.
    """

    def __init__(self, host: str = "0.0.0.0", port: int = 8000):
        super().__init__()
        self.host = host
        self.port = port
        self.app = FastAPI()
        self.client: Optional[httpx.AsyncClient] = None
        self.active_connections: List[WebSocket] = []
        self._setup_app()

    def _setup_app(self):
        """Configure FastAPI routes and event handlers."""
        self.app.state.etl_server = self
        self.app.add_event_handler("startup", self.startup_event)
        self.app.add_event_handler("shutdown", self.shutdown_event)

        @self.app.get("/health")
        async def health_check():
            return Response(content=b"Running")

        @self.app.get("/{path:path}")
        async def handle_get(path: str, request: Request):
            return await self._handle_request(path, request, is_get=True)

        @self.app.put("/{path:path}")
        async def handle_put(path: str, request: Request):
            return await self._handle_request(path, request, is_get=False)

        @self.app.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            try:
                await websocket.accept()
                self.logger.debug(
                    "WebSocket connection established: %s", websocket.client
                )
                self.active_connections.append(websocket)

                while True:
                    await self._handle_ws_message(websocket)

            except Exception as e:
                self.logger.error(
                    "Unexpected WebSocket error from %s: %s", websocket.client, e
                )
            finally:
                if websocket in self.active_connections:
                    self.active_connections.remove(websocket)
                try:
                    await websocket.close()  # might have already been closed by peer
                except RuntimeError as e:
                    self.logger.debug("Skip close: %s", e)

    async def startup_event(self):
        """Initialize resources on server startup."""
        self.client = httpx.AsyncClient(timeout=None, limits=HTTP_LIMITS)
        self.logger.info("Server starting up")

    async def shutdown_event(self):
        """Cleanup resources on server shutdown."""
        await self.client.aclose()
        self.logger.info("Server shutting down")

    # pylint: disable=too-many-locals
    async def _handle_request(self, path: str, request: Request, is_get: bool):
        """Unified request handler for GET/PUT operations."""
        self.logger.debug(
            "Processing %s request for path: %s", "GET" if is_get else "PUT", path
        )
        etl_args = request.query_params.get(QPARAM_ETL_ARGS, "").strip()
        fqn = request.query_params.get(QPARAM_ETL_FQN, "").strip()
        self.logger.debug("etl_args = %r, fqn = %r", etl_args, fqn)

        try:
            if fqn:
                content = await self._get_fqn_content(fqn)
            elif is_get:
                content = await self._get_network_content(path)
            else:
                content = await request.body()

            # Transform the content
            transformed = self.transform(content, path, etl_args)

            # Handle pipeline if present
            pipeline_header = request.headers.get(HEADER_NODE_URL)
            self.logger.debug("pipeline_header: %r", pipeline_header)
            if pipeline_header:
                first_url, remaining_pipeline = parse_etl_pipeline(pipeline_header)
                if first_url:
                    status_code, transformed, direct_put_length = (
                        await self._direct_put(
                            first_url, transformed, remaining_pipeline, path
                        )
                    )
                    self.logger.debug("status_code: %r", status_code)

                    return Response(
                        content=transformed,
                        status_code=status_code,
                        headers=(
                            {HEADER_DIRECT_PUT_LENGTH: str(direct_put_length)}
                            if direct_put_length != 0
                            else {}
                        ),
                    )

            self.logger.debug(
                "no pipeline, returning transformed content directly, length: %r",
                len(transformed),
            )
            # No pipeline, return transformed content directly
            return Response(
                content=transformed,
                status_code=STATUS_OK,
                media_type=self.get_mime_type(),
            )

        except InvalidPipelineError as e:
            self.logger.error("Invalid pipeline header: %s", str(e))
            raise HTTPException(
                status_code=400, detail=f"Invalid pipeline header: {str(e)}"
            ) from e
        except FileNotFoundError as exc:
            fs_path = exc.filename or path
            self.logger.error(
                "Error processing object %r: file not found at %r",
                path,
                fs_path,
            )
            raise HTTPException(
                404,
                detail=(
                    f"Error processing object {path!r}: file not found at {fs_path!r}."
                ),
            ) from exc
        except httpx.HTTPStatusError as e:
            self.logger.warning(
                "Target responded with error: %s", e.response.status_code
            )
            raise HTTPException(
                e.response.status_code, detail="Target request failed"
            ) from e
        except httpx.RequestError as e:
            self.logger.error("Network error: %s", str(e))
            raise HTTPException(502, detail=f"Network error: {str(e)}") from e
        except Exception as e:
            self.logger.exception("Critical error during processing")
            raise HTTPException(500, detail=f"Processing error: {str(e)}") from e

    async def _get_fqn_content(self, path: str) -> bytes:
        """Safely read local file content with path normalization."""
        decoded_path = unquote(path)
        safe_path = os.path.normpath(os.path.join("/", decoded_path.lstrip("/")))
        self.logger.debug("Reading local file: %s", safe_path)

        async with aiofiles.open(safe_path, "rb") as f:
            return await f.read()

    async def _get_network_content(self, path: str) -> bytes:
        """Retrieve content from AIS target with async HTTP client."""
        obj_path = quote(path, safe="@")
        target_url = f"{self.host_target}/{obj_path}"
        self.logger.debug("Forwarding to target: %s", target_url)

        response = await self.client.get(target_url)
        response.raise_for_status()
        return response.content

    async def _direct_put(
        self,
        direct_put_url: str,
        data: bytes,
        remaining_pipeline: str = "",
        path: str = "",
    ) -> Tuple[int, bytes, int]:
        """
        Sends the transformed object directly to the specified AIS node (`direct_put_url`),
        eliminating the additional network hop through the original target.
        Used only in bucket-to-bucket offline transforms.

        Args:
            direct_put_url: The first URL in the ETL pipeline
            data: The transformed data to send
            remaining_pipeline: Comma-separated remaining pipeline stages to pass as header
            path: The path of the object.
        Returns:
            status code, transformed data, length of the transformed data (if any)
        """
        try:
            url = compose_etl_direct_put_url(direct_put_url, self.host_target, path)
            headers = {}
            if remaining_pipeline:
                headers[HEADER_NODE_URL] = remaining_pipeline
            # TODO: add etl_args to qparams if present

            resp = await self.client.put(url, content=data, headers=headers)
            return self.handle_direct_put_response(resp, data)

        except Exception as exc:
            error = str(exc).encode()
            self.logger.error("Direct put exception to %s: %s", direct_put_url, exc)
            return STATUS_INTERNAL_SERVER_ERROR, error, 0

    def _build_response(self, content: bytes, mime_type: str) -> Response:
        """Construct standardized response with appropriate headers."""
        return Response(
            content=content,
            media_type=mime_type,
            headers={HEADER_CONTENT_LENGTH: str(len(content))},
        )

    async def _handle_ws_message(self, websocket: WebSocket):
        """Handle a single WebSocket message."""
        ctrl_msg = await websocket.receive_json(mode="binary")
        self.logger.debug("Received control message: %s", ctrl_msg)

        fqn = ctrl_msg.get(ETL_WS_FQN)
        path = ctrl_msg.get(ETL_WS_PATH)
        content = (
            await self._get_fqn_content(fqn) if fqn else await websocket.receive_bytes()
        )
        etl_args = ctrl_msg.get(QPARAM_ETL_ARGS)

        try:
            transformed = await asyncio.to_thread(
                self.transform, content, path, etl_args
            )

            pipeline_header = ctrl_msg.get(ETL_WS_PIPELINE)
            if pipeline_header:
                self.logger.debug("pipeline_header: %r", pipeline_header)
                first_url, remaining_pipeline = parse_etl_pipeline(pipeline_header)
                if first_url:
                    status_code, transformed, direct_put_length = (
                        await self._direct_put(
                            first_url, transformed, remaining_pipeline, path
                        )
                    )
                    if status_code == STATUS_OK:
                        await websocket.send_bytes(transformed)
                    else:
                        await websocket.send_text(str(direct_put_length))
                    return

            # No pipeline, send transformed data
            await websocket.send_bytes(transformed)

        except InvalidPipelineError as e:
            self.logger.error("Invalid pipeline header: %s", str(e))
            await websocket.send_text(f"Invalid pipeline header: {str(e)}")
        except Exception as e:
            self.logger.error("Transform error: %s", str(e))
            await websocket.send_text(f"Transform error: {str(e)}")

    def start(self):
        """Start the server with production-optimized settings."""
        uvicorn.run(self.app, host=self.host, port=self.port)



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from dateutil.parser import isoparse

from aistore.sdk.types import JobSnap


@dataclass
class WaitResult:
    """
    Result of waiting for a job to reach a desired state.

    Attributes:
        job_id: The job ID
        success: True if the wait condition was met successfully
        error: Error message, if any
        end_time: When the job ended, if any
    """

    job_id: str
    success: bool
    error: Optional[str] = None
    end_time: Optional[datetime] = None

    def __bool__(self) -> bool:
        """Returns True if wait operation was successful."""
        return self.success

    @classmethod
    def from_snapshots(cls, job_id: str, snapshots: List[JobSnap]):
        """
        Create a WaitResult from a list of job snapshots.

        Args:
            job_id: The job ID
            snapshots: List of job snapshots from all target nodes

        Returns:
            WaitResult indicating success/failure based on snapshot states
        """
        aborted = False
        error_msg = None
        all_ended = True
        end_times = []

        for s in snapshots:
            if s.aborted:
                aborted = True
            if s.abort_err and error_msg is None:
                error_msg = s.abort_err
            if not (s.end_time and s.end_time != "0001-01-01T00:00:00Z"):
                all_ended = False
            elif s.end_time:
                end_times.append(isoparse(s.end_time))

        success = not aborted and not error_msg
        error = error_msg or ("aborted" if aborted else None)
        end_time = max(end_times) if all_ended and end_times else None

        return cls(job_id=job_id, success=success, error=error, end_time=end_time)



#
# Copyright (c) 2022-2026, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations  # pylint: disable=unused-variable

import logging
from typing import Dict, List, Optional, Union

from aistore.sdk.const import (
    HTTP_METHOD_GET,
    ACT_LIST,
    QPARAM_WHAT,
    QPARAM_PRIMARY_READY_REB,
    QPARAM_PROVIDER,
    HEADER_NODE_ID,
    URL_PATH_ETL,
    URL_PATH_REVERSE,
    URL_PATH_BUCKETS,
    URL_PATH_HEALTH,
    URL_PATH_DAEMON,
    URL_PATH_CLUSTER,
    WHAT_SMAP,
    WHAT_ALL_XACT_STATUS,
    WHAT_ALL_RUNNING_STATUS,
    WHAT_NODE_STATS_AND_STATUS,
)
from aistore.sdk.etl.etl_const import ETL_STAGE_RUNNING
from aistore.sdk.provider import Provider

from aistore.sdk.types import (
    BucketModel,
    JobStatus,
    JobQuery,
    ETLInfo,
    ActionMsg,
    Smap,
)
from aistore.sdk.request_client import RequestClient

logger = logging.getLogger("cluster")


# pylint: disable=unused-variable
class Cluster:
    """
    A class representing a cluster bound to an AIS client.
    """

    # pylint: disable=duplicate-code
    def __init__(self, client: RequestClient):
        self._client = client

    @property
    def client(self):
        """Client this cluster uses to make requests"""
        return self._client

    def get_info(self) -> Smap:
        """
        Returns state of AIS cluster, including the detailed information about its nodes.

        Returns:
            aistore.sdk.types.Smap: Smap containing cluster information

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        return self._get_smap()

    def get_primary_url(self) -> str:
        """
        Returns: URL of primary proxy
        """
        return self._get_smap().proxy_si.public_net.direct_url

    def list_buckets(self, provider: Union[str, Provider] = Provider.AIS):
        """
        Returns list of buckets in AIStore cluster.

        Args:
            provider (str or Provider, optional): Provider of bucket (one of "ais", "aws", "gcp", ...).
                Defaults to "ais". Empty provider returns buckets of all providers.

        Returns:
            List[BucketModel]: A list of buckets

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        params = {QPARAM_PROVIDER: Provider.parse(provider).value}
        action = ActionMsg(action=ACT_LIST).model_dump()

        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_BUCKETS,
            res_model=List[BucketModel],
            json=action,
            params=params,
        )

    def list_jobs_status(self, job_kind="", target_id="") -> List[JobStatus]:
        """
        List the status of jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of JobStatus objects
        """
        res = self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=Optional[List[JobStatus]],
            json=JobQuery(kind=job_kind, target=target_id).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_XACT_STATUS},
        )
        if res is None:
            return []
        return res

    def list_running_jobs(self, job_kind="", target_id="") -> List[str]:
        """
        List the currently running jobs on the cluster

        Args:
            job_kind (str, optional): Only show jobs of a particular type
            target_id (str, optional): Limit to jobs on a specific target node

        Returns:
            List of jobs in the format job_kind[job_id]
        """
        return self._client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_CLUSTER,
            res_model=List[str],
            json=JobQuery(kind=job_kind, target=target_id, active=True).as_dict(),
            params={QPARAM_WHAT: WHAT_ALL_RUNNING_STATUS},
        )

    def list_etls(self, stages: Optional[List[str]] = None) -> List[ETLInfo]:
        """
        Lists ETLs filtered by their stages.

        Args:
            stages (List[str], optional): List of stages to filter ETLs by. Defaults to ["running"].

        Returns:
            List[ETLInfo]: A list of details on ETLs matching the specified stages
        """
        if stages is None:
            stages = [ETL_STAGE_RUNNING]

        res = self._client.request_deserialize(
            HTTP_METHOD_GET, path=URL_PATH_ETL, res_model=List[ETLInfo]
        )
        return [info for info in res if info.stage in stages]

    def is_ready(self) -> bool:
        """
        Checks if cluster is ready or still setting up.

        Returns:
            bool: True if cluster is ready, or false if cluster is still setting up
        """
        # compare with AIS Go API (api/cluster.go) for additional supported options
        params = {QPARAM_PRIMARY_READY_REB: "true"}
        try:
            resp = self._client.request(
                HTTP_METHOD_GET,
                path=URL_PATH_HEALTH,
                endpoint=self.get_primary_url(),
                params=params,
            )
            return resp.ok
        except Exception as err:
            logger.debug(err)
            return False

    def get_performance(self) -> Dict:
        """
        Retrieves the raw performance and status data from each target node in the AIStore cluster.

        Returns:
            Dict: A dictionary where each key is the ID of a target node and each value is the
                    raw AIS performance/status JSON returned by that node (for more information,
                    see https://aistore.nvidia.com/docs/monitoring-metrics#target-metrics).

        Raises:
            requests.RequestException: If there's an ambiguous exception while processing the request
            requests.ConnectionError: If there's a connection error with the cluster
            requests.ConnectionTimeout: If the connection to the cluster times out
            requests.ReadTimeout: If the timeout is reached while awaiting a response from the cluster
        """
        performance_data = {}
        targets = self._get_targets()
        params = {QPARAM_WHAT: WHAT_NODE_STATS_AND_STATUS}

        for target_id in targets:
            headers = {HEADER_NODE_ID: target_id}
            resp = self.client.request(
                HTTP_METHOD_GET,
                path=f"{URL_PATH_REVERSE}/{URL_PATH_DAEMON}",
                params=params,
                headers=headers,
            )
            resp.raise_for_status()
            performance_data[target_id] = resp.json()

        return performance_data

    def _get_smap(self):
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_DAEMON,
            res_model=Smap,
            params={QPARAM_WHAT: WHAT_SMAP},
        )

    def _get_targets(self):
        tmap = self._get_smap().tmap
        return list(tmap.keys())

    def get_uuid(self) -> str:
        """
        Returns: UUID of AIStore Cluster
        """
        return self._get_smap().uuid



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from typing import Optional, Iterator, Tuple, Union
from re import search, IGNORECASE
from requests.models import Response

from aistore.sdk.utils import get_logger
from aistore.sdk.const import (
    MULTIPART_MARKER,
    WIN_LINE_END,
    UNIX_LINE_END,
    UTF_ENCODING,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_MAX_BUFFER_SIZE,
    HTTP_BOUNDARY_REGEX,
)
from aistore.sdk.batch.multipart.stateful_streaming_parser import (
    StatefulStreamingParser,
)
from aistore.sdk.batch.errors import MultipartDecodeError
from aistore.sdk.batch.multipart.body_stream_reader import BodyStreamReader

logger = get_logger(__name__)


# pylint: disable=too-few-public-methods
class MultipartDecoder:
    """
    A multipart decoder for parsing multipart HTTP responses.

    Handles boundary extraction, content parsing, and data extraction from individual parts.
    The decoder supports both Windows (\\r\\n) and Unix (\\n) line endings and gracefully
    handles malformed content by skipping invalid parts.

    Args:
        encoding (str, optional): Character encoding to use. Defaults to "utf-8".
        parse_as_stream (bool, optional): If True, yields part body content as
            an iterator of chunks that can be directly read from. If False,
            yields entire content in memory as bytes. Defaults to False. Note that
            this functionality is in development, not fully tested, and subject to change.
        chunk_size (int, optional): Size of chunks when streaming. Defaults to 8192.
        max_buffer_size (int, optional): Maximum buffer size for sliding window. Defaults to 64KB.

    Example:
        >>> # Traditional usage (loads entire body into memory)
        >>> decoder = MultipartDecoder()
        >>> for headers, body in decoder.decode(response):
        ...     print(f"Headers: {headers}, Body size: {len(body)} bytes")

        >>> # Streaming usage (processes body in chunks)
        >>> decoder = MultipartDecoder(parse_as_stream=True)
        >>> for headers, body_stream in decoder.decode(response):
        ...     for chunk in body_stream:
        ...         process_chunk(chunk)
    """

    def __init__(
        self,
        encoding: str = UTF_ENCODING,
        parse_as_stream: bool = False,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        max_buffer_size: int = DEFAULT_MAX_BUFFER_SIZE,
    ):
        """
        Initialize MultipartDecoder with specified options.

        Args:
            encoding (str, optional): Character encoding to use. Defaults to "utf-8".
            parse_as_stream (bool, optional): Enable parsing as stream mode. Defaults to False.
            chunk_size (int, optional): Size of chunks when streaming. Defaults to 32KiB.
            max_buffer_size (int, optional): Maximum buffer size for sliding window. Defaults to 64KiB.
        """
        self.encoding = encoding
        self.parse_as_stream = parse_as_stream
        self.chunk_size = chunk_size
        self.max_buffer_size = max_buffer_size

    def decode(
        self, response: Response
    ) -> Iterator[Tuple[bytes, Union[bytes, BodyStreamReader]]]:
        """
        Decode contents of a multipart HTTP response.

        Args:
            response (Response): HTTP response object containing multipart data

        Yields:
            Tuple[bytes, Union[bytes, BodyStreamReader]]: Extracted body parts.
                If parse_as_stream=False, yields (headers, body).
                If parse_as_stream=True, yields (headers, body_stream).
        """
        # Validate content type
        content_type = response.headers.get("Content-Type", "")
        if "multipart" not in content_type.lower():
            raise MultipartDecodeError("Response is not of multipart content type")

        # Extract boundary from content type header
        # Match boundary format: boundary=<BOUNDARY-ID>
        match = search(HTTP_BOUNDARY_REGEX, content_type, IGNORECASE)

        if match is None:
            raise ValueError("Boundary not found in Content-Type header")

        # Remove quotes if present
        boundary = match.group(1).strip("\"'")

        if self.parse_as_stream:
            yield from self._parse_content_streaming(response, boundary)
        else:
            yield from self._parse_content(response.content, boundary)

    def _parse_content(
        self,
        content: bytes,
        boundary: str,
    ) -> Iterator[Tuple[bytes, bytes]]:
        """
        Parse multipart content and extract data from each body part (non-streaming).

        Args:
            boundary (str): Boundary identifier used to separate body parts
            content (bytes): Raw multipart content to parse

        Returns:
            Iterator[Tuple[bytes, bytes]]: Iterator over extracted body part header and data
        """
        # Create delimiter: -- + boundary
        delimiter = MULTIPART_MARKER + boundary.encode(self.encoding)

        # Split content by delimiter
        # Skip first part (preamble) and process remaining parts
        for part in content.split(delimiter)[1:]:
            # Skip empty parts or end markers
            if not part.strip() or part.strip() == MULTIPART_MARKER:
                continue

            # Extract data from part (headers and body part)
            part_data = self._parse_part(part)
            if part_data:
                yield part_data

    def _parse_part(self, part: bytes) -> Optional[Tuple[bytes, bytes]]:
        """
        Parse individual multipart body part and extract headers and body.

        Args:
            part (bytes): Raw body part content including headers and body

        Returns:
            Optional[Tuple[bytes, bytes]]: Extracted headers and body, or None if parsing fails
        """
        # Determine line ending style
        if WIN_LINE_END in part:
            line_ending = WIN_LINE_END
        elif UNIX_LINE_END in part:
            line_ending = UNIX_LINE_END
        else:
            # No valid line ending found
            logger.warning(
                "Multipart sections must include either Windows-style line endings %s "
                "or Unix-style line endings %s. The part has ending: %s",
                WIN_LINE_END,
                UNIX_LINE_END,
                part or "",
            )
            return None

        # Split headers from body
        try:
            # Headers and body are separated by double line ending
            sections = part.split(line_ending, 1)

            # If splitting failed, then part is malformed
            if not sections or len(sections) == 1:
                logger.exception(
                    "Malformed part: expected header/body split\n%s",
                    part.decode(self.encoding, errors="replace"),
                )
                return None

            headers, body = sections
            return headers, body

        except (ValueError, IndexError):
            # Failed to parse part
            logger.exception(
                "Could not parse part by splitting headers from body \n%s",
                str(part),
            )
            return None

    def _parse_content_streaming(
        self, response: Response, boundary: str
    ) -> Iterator[Tuple[bytes, BodyStreamReader]]:
        """
        Use a streaming parser to start decoding incoming multipart stream.
        Only extracts headers in memory and streams the body content per part.
        """
        start_boundary = MULTIPART_MARKER + boundary.encode(self.encoding)

        # Create stateful parser using boundaries and chunk size
        state_parser = StatefulStreamingParser(
            response.iter_content(chunk_size=self.chunk_size),
            start_boundary,
            self.max_buffer_size,
        )

        # Yield each part as we encounter it
        while True:
            part_result = state_parser.get_next_part()
            if part_result is None:
                break

            # headers and body stream
            yield part_result



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import zipfile
from typing import Any, Generator, Optional, Tuple, Union
from io import BytesIO

from overrides import override
from requests import Response

from aistore.sdk.batch.types import MossOut, MossReq, MossResp
from aistore.sdk.batch.extractor.archive_stream_extractor import ArchiveStreamExtractor
from aistore.sdk.const import EXT_ZIP
from aistore.sdk.utils import get_logger

logger = get_logger(__name__)


class ZipStreamExtractor(ArchiveStreamExtractor):
    """
    Class for extracting batch .zip streams from AIStore.

    Integrates with MossReq/MossResp objects to provide proper metadata mapping.
    """

    _supported_fmts = (EXT_ZIP,)

    @override
    def extract(
        self,
        response: Response,
        data_stream: Union[BytesIO, Any],
        moss_req: MossReq,
        moss_resp: Optional[MossResp] = None,
    ) -> Generator[Tuple[MossOut, bytes], None, None]:
        """
        Extract from zip archive stream.

        Note: ZIP format requires random access, so the entire stream
        is loaded into memory before extraction begins.

        Args:
            response (Response): HTTP response
            data_stream (Union[BytesIO, Any]): Archive data
            moss_req (MossReq): Original request
            moss_resp (Optional[MossResp]): Response metadata (None for streaming)

        Yields:
            Tuple[MossOut, bytes]: (MossOut, content) tuples
        """
        # ZIP requires random access - load into memory if needed
        if not hasattr(data_stream, "read"):
            data_stream = BytesIO(data_stream)
        elif moss_req.streaming_get:
            # For streaming mode, read entire response into memory
            data_stream = BytesIO(data_stream.read())

        index = 0
        try:
            with zipfile.ZipFile(data_stream, "r") as zip_file:
                for zip_info in zip_file.infolist():
                    if zip_info.is_dir():
                        continue

                    try:
                        content = zip_file.read(zip_info.filename)

                        # Get MossOut (from response or build from request)
                        moss_out = self._get_moss_out(
                            index, len(content), moss_req, moss_resp
                        )

                        # Check for missing
                        # TODO: FIXME: what to do with missing files?
                        # moss_out.is_missing = zip_info.filename.startswith(GB_MISSING_FILES_DIR)

                        index += 1
                        yield moss_out, content

                    except (
                        zipfile.BadZipFile,
                        zipfile.LargeZipFile,
                        KeyError,
                        OSError,
                    ) as e:
                        # Handle individual file extraction errors
                        self._handle_extraction_error(
                            zip_info.filename, e, moss_req, EXT_ZIP
                        )
                        index += 1
                        continue

        except (zipfile.BadZipFile, zipfile.LargeZipFile, OSError) as e:
            # Handle zip stream errors (corrupt archive, etc.)
            logger.error("Failed to read zip archive: %s", str(e))
            raise RuntimeError("Failed to read zip archive stream") from e
        finally:
            response.close()



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import base64
import binascii
from typing import Optional, List, Dict

from pydantic import BaseModel, Field, field_validator

from aistore.sdk.const import (
    GB_ERR_MSG,
    GB_OBJNAME,
    GB_BCK,
    GB_OUT,
    GB_PROVIDER,
    GB_UNAME,
    GB_OPAQUE,
    GB_ARCHPATH,
    GB_START,
    GB_LENGTH,
    GB_IN,
    GB_OUTPUT_FMT,
    GB_CONTINUE_ERR,
    GB_ONLY_OBJ_NAME,
    GB_STRM_GET,
    GB_COLOC,
    GB_SIZE,
    GB_UUID,
)


# pylint: disable=too-few-public-methods
class MossIn(BaseModel):
    """Matches apc.MossIn in Go."""

    obj_name: str = Field(alias=GB_OBJNAME)  # name of the object
    # Optional fields
    bck: Optional[str] = Field(
        default=None, alias=GB_BCK
    )  # if present, overrides cmn.Bck from the GetBatch request
    provider: Optional[str] = Field(
        default=None, alias=GB_PROVIDER
    )  # e.g. "s3", "ais", etc.
    uname: Optional[str] = Field(
        default=None, alias=GB_UNAME
    )  # per-object, fully qualified - defines the entire (bucket, provider, objname) triplet, and more
    opaque: Optional[str] = Field(
        default=None, alias=GB_OPAQUE
    )  # user-provided identifier - e.g., to maintain one-to-many
    archpath: Optional[str] = Field(
        default=None, alias=GB_ARCHPATH
    )  # extract the specified file from an object ("shard") formatted as: .tar, .tgz or .tar.gz, .zip, .tar.lz4;
    start: Optional[int] = Field(
        default=None, alias=GB_START
    )  # start offset in the object
    length: Optional[int] = Field(
        default=None, alias=GB_LENGTH
    )  # length of the data to be extracted

    model_config = {"populate_by_name": True}

    def dict(self, *_args, **_kwargs) -> Dict:
        return self.model_dump(by_alias=True, exclude_defaults=True)


# pylint: disable=too-few-public-methods
class MossOut(BaseModel):
    """MOSS (Multi-Object Streaming Service) response. Matches apc.MossOut in Go - Response for each object"""

    obj_name: str = Field(alias=GB_OBJNAME)  # name of the object
    archpath: Optional[str] = Field(
        default=None, alias=GB_ARCHPATH
    )  # path of the object in the shard
    bucket: str = Field(alias=GB_BCK)
    provider: str = Field(alias=GB_PROVIDER)
    opaque: Optional[bytes] = Field(default=None, alias=GB_OPAQUE)
    err_msg: Optional[str] = Field(default=None, alias=GB_ERR_MSG)
    size: int = Field(default=0, alias=GB_SIZE)

    model_config = {"populate_by_name": True}

    def dict(self, *_args, **_kwargs) -> Dict:
        return self.model_dump(by_alias=True, exclude_defaults=True)

    # pylint: disable=no-self-argument
    @field_validator("opaque", mode="before")
    def decode_opaque(cls, opaque_val):
        """
        Automatically decode base64 encoded opaque field back to bytes.

        Args:
            opaque_val (Union[str, bytes, None]): The raw opaque value. Can be a base64-encoded
                string, already decoded bytes, or None if no opaque data was provided.

        Returns:
            Optional[bytes]: Decoded bytes or None
        """
        if opaque_val is None:
            return None
        # Already decoded
        if isinstance(opaque_val, bytes):
            return opaque_val
        if isinstance(opaque_val, str) and opaque_val:
            try:
                return base64.urlsafe_b64decode(opaque_val)
            except (binascii.Error, ValueError):
                # Handle invalid base64 gracefully
                return None
        return None


# pylint: disable=too-few-public-methods
class MossReq(BaseModel):
    """MOSS (Multi-Object Streaming Service) request. Matches apc.MossReq in Go."""

    moss_in: List[MossIn] = Field(default_factory=list, alias=GB_IN)  # MossIn Requests
    # Optional fields
    output_format: Optional[str] = Field(
        default=None, alias=GB_OUTPUT_FMT
    )  # tar, tgz, zip, etc.
    cont_on_err: Optional[bool] = Field(
        default=None, alias=GB_CONTINUE_ERR
    )  # primary usage: ignore missing files and/or objects - include them under "__404__/" prefix and keep going
    only_obj_name: Optional[bool] = Field(
        default=None, alias=GB_ONLY_OBJ_NAME
    )  # name-in-archive: default naming convention is <Bucket>/<ObjName>; set this flag to have <ObjName> only
    streaming_get: Optional[bool] = Field(
        default=None, alias=GB_STRM_GET
    )  # stream resulting archive prior to finalizing it in memory
    colocation: Optional[int] = Field(
        default=None, alias=GB_COLOC
    )  # colocation hint: 0=none, 1=target-aware, 2=target+shard-aware

    model_config = {"populate_by_name": True}

    def dict(self, *_args, **_kwargs) -> Dict:
        return self.model_dump(by_alias=True, exclude_defaults=True)

    def add(self, moss_in: MossIn) -> "MossReq":
        """
        Add a MossIn request. Returns self for chaining.

        Args:
            moss_in (MossIn): MossIn object to add to the request

        Returns:
            MossReq: Self for method chaining
        """
        self.moss_in.append(moss_in)  # pylint: disable=no-member
        return self

    def extend(self, moss_ins: List[MossIn]) -> "MossReq":
        """
        Add multiple MossIn requests. Returns self for chaining.

        Args:
            moss_ins (List[MossIn]): List of MossIn objects to add

        Returns:
            MossReq: Self for method chaining
        """
        self.moss_in.extend(moss_ins)  # pylint: disable=no-member
        return self


# pylint: disable=too-few-public-methods
class MossResp(BaseModel):
    """Matches apc.MossResp in Go"""

    out: List[MossOut] = Field(default_factory=list, alias=GB_OUT)
    uuid: str = Field(default="", alias=GB_UUID)

    model_config = {"populate_by_name": True}



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from typing import Optional, Iterator, Tuple

from aistore.sdk.const import WIN_LINE_END, UNIX_LINE_END
from aistore.sdk.utils import get_logger
from aistore.sdk.batch.multipart.multipart_stream_buffer import MultipartStreamBuffer
from aistore.sdk.batch.multipart.body_stream_reader import BodyStreamReader

logger = get_logger(__name__)


class StatefulStreamingParser:
    """
    A stateful streaming parser for efficiently processing multipart HTTP responses.

    The parser maintains state across multiple parts of a multipart response,
    automatically detecting boundaries, extracting headers, and providing streaming
    access to body content.

    Args:
        content_iter (Iterator[bytes]): Iterator yielding chunks of the HTTP response
        boundary (bytes): The multipart boundary marker (including leading dashes)
        max_buffer_size (int): Maximum size of the internal sliding window buffer

    Notes:
        This class is not thread-safe. Each instance should be used by a single
        thread or proper synchronization must be implemented by the caller.
    """

    def __init__(
        self,
        content_iter: Iterator[bytes],
        boundary: bytes,
        max_buffer_size: int,
    ):
        self._boundary = boundary
        self._max_buffer_size = max_buffer_size

        # Initialize buffer reader
        self._stream_buffer = MultipartStreamBuffer(
            content_iter, boundary, max_buffer_size
        )

        self._first_boundary_found = False

        # Line ending in response
        self._cached_line_end = None

    def get_next_part(self) -> Optional[Tuple[bytes, BodyStreamReader]]:
        """
        Get the next part without loading it completely into memory.

        Returns:
            Optional[Tuple[bytes, BodyStreamReader]]: If part exists, returns tuple of headers and body stream
        """
        if self._stream_buffer.is_exhausted():
            return None

        # Find and skip to first boundary if not found yet
        if not self._first_boundary_found:
            if not self._locate_first_boundary():
                return None

        # Extract headers for current part
        headers = self._extract_headers()
        if headers is None:
            return None

        # Create body stream for the rest of this part
        body_stream = BodyStreamReader(self._stream_buffer)

        return headers, body_stream

    def _locate_first_boundary(self) -> bool:
        """
        Find the first boundary marker using the optimized buffer reader.

        Returns:
            bool: True if boundary found, False if exhausted without finding boundary
        """
        min_size = len(self._boundary) * 2

        while self._stream_buffer.has_data():
            # Use the optimized ensure_data_available method
            if not self._stream_buffer.ensure_data_available(min_size):
                return False

            # Find boundary position
            boundary_pos = self._stream_buffer.find_pattern(self._boundary)
            if boundary_pos != -1:
                # Consume everything up to and including the boundary
                self._stream_buffer.consume_data(
                    boundary_pos + self._stream_buffer.get_boundary_size()
                )

                # Handle boundary-specific cleanup
                self._stream_buffer.skip_whitespace()

                self._first_boundary_found = True
                return True
            break

        return False

    def _extract_headers(self) -> Optional[bytes]:
        """
        Extract headers using the optimized content reader abstraction.

        Returns:
            Optional[bytes]: Headers as bytes if found, None if stream exhausted
        """
        while self._stream_buffer.has_data():
            headers_end, detected_line_end = -1, None

            # Single ensure_data_available call
            self._stream_buffer.ensure_data_available()

            for line_end in (WIN_LINE_END, UNIX_LINE_END):
                pos = self._stream_buffer.find_pattern(line_end)
                if pos != -1:
                    headers_end, detected_line_end = pos, line_end

            if headers_end != -1:
                # Extract headers
                headers = self._stream_buffer.get_data_slice(0, headers_end)

                # Cache the detected line ending (no redundant buffer scan)
                if not self._cached_line_end:
                    self._cached_line_end = detected_line_end

                # Consume headers + line ending
                self._stream_buffer.consume_data(
                    headers_end + len(self._cached_line_end)
                )
                return headers

            if not self._stream_buffer.ensure_data_available():
                break

        return None



#
# Copyright (c) 2023-2025, NVIDIA CORPORATION. All rights reserved.
#
from typing import Callable, Optional

import requests
import tenacity
from requests import Response

from aistore.sdk.retry_config import RetryConfig
from aistore.sdk.const import HTTP_METHOD_GET, URL_PATH_OBJECTS
from aistore.sdk.presence_poller import PresencePoller
from aistore.sdk.provider import Provider
from aistore.sdk import utils

logger = utils.get_logger(__name__)


# pylint: disable=too-few-public-methods
class RetryManager:
    """
    Manages applying the retry configuration along with any custom logic to requests.

    Args:
        request_func (Callable): A function that takes request parameters and returns a requests.Response object.
        retry_config (Optional[RetryConfig]): Retry configuration, containing config for both tenacity retry logic
            and logic for cold gets from remote backend providers.
    """

    def __init__(
        self,
        request_func: Callable[..., Response],
        retry_config: Optional[RetryConfig] = None,
    ):
        self._retry_config = retry_config or RetryConfig.default()
        self._presence_poller = PresencePoller(
            request_func, self._retry_config.cold_get_conf
        )
        self._retrying = self._retry_config.network_retry.copy()
        # In case the provided Retrying instance already has a before_sleep, append our custom one first
        if (
            retry_config
            and retry_config.network_retry
            and retry_config.network_retry.before_sleep
        ):

            def combined_before_sleep(*args):
                self._before_sleep(*args)
                retry_config.network_retry.before_sleep(*args)

            self._retrying.before_sleep = combined_before_sleep
        else:
            self._retrying.before_sleep = self._before_sleep

    @property
    def retry_config(self) -> RetryConfig:
        """Return the retry config provided to this manager"""
        return self._retry_config

    def _before_sleep(self, retry_state: tenacity.RetryCallState):
        """
        Hook called by Tenacity before sleeping between retries.
        Determine if we want to delay the retry.
        If so, use PresencePoller to poll until the object is present.

        Args:
            retry_state (tenacity.RetryCallState): Retry state from tenacity.
        """
        outcome = retry_state.outcome
        if not outcome:
            return
        exc = retry_state.outcome.exception()

        # Early return if no special handling is needed
        if not isinstance(exc, requests.ConnectionError):
            return
        if not self._should_delay_retry(exc):
            return
        # Otherwise: poll for object presence
        logger.info(
            "Delaying retry by polling for object presence after: '%s'",
            type(exc).__name__,
        )
        try:
            self._presence_poller.wait_for_presence(exc.request)
        except Exception as retry_err:
            # Allow tenacity to continue retrying and do not re-raise
            logger.error("Error while polling for object presence: %s", retry_err)

    @staticmethod
    def _should_delay_retry(exc: requests.ConnectionError) -> bool:
        """
        If we get a read timeout, it's possible AIS is currently downloading and writing a remote object,
            so our initial request failed to acquire a read lock.

        Retrying immediately means we expect AIS to be finished downloading from remote,
            which may take longer than our default retry.

        Args:
            exc (requests.ConnectionError): Original exception.
        Returns:
            Whether to delay tenacity retry.
        """
        if not utils.is_read_timeout(exc):
            return False
        if exc.request is None:
            return False
        req = exc.request
        exc_type = type(exc).__name__
        # Do nothing if it's not a GET
        if req.method is None or req.method.lower() != HTTP_METHOD_GET:
            logger.debug("Received error: '%s' from non-GET request", exc_type)
            return False
        # Do nothing if it's not a request for an object
        if f"{URL_PATH_OBJECTS}/" not in req.url:
            logger.debug("Received error: '%s' from non-object request", exc_type)
            return False
        # Do nothing if the bucket is not remote
        provider = utils.get_provider_from_request(req)
        if provider == Provider.AIS:
            logger.debug(
                "Received error: '%s' from object request to AIS bucket", exc_type
            )
            return False
        logger.debug(
            "Received error: '%s' from object GET request to bucket with provider '%s'",
            exc_type,
            provider.value,
        )
        return True

    def with_retry(
        self, request_func: Callable[..., Response], *args, **kwargs
    ) -> Response:
        """
        **Why `tenacity.retry` over `urllib3.Retry`?**
        `urllib3.Retry` always retries the **same failing URL**, which is problematic if a target is down or restarting.
        Instead, we retry at this stage with tenacity to re-create the initial request to the **proxy URL**.
        This request will then get redirected to the latest selected target.

        Args:
            request_func (Callable): Function that makes a request to AIS and returns a requests.Response object.
            *args: Args for request_func.
            **kwargs: Kwargs for request_func.

        Returns:
            Response object from the provided request_func.
        """
        return self._retrying(request_func, *args, **kwargs)



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from io import BufferedIOBase
from overrides import override

from aistore.sdk.batch.multipart.multipart_stream_buffer import MultipartStreamBuffer


class BodyStreamReader(BufferedIOBase):
    """
    Stream reader that only reads the body portion of a part from a
    buffered multipart stream. This class extends BufferedIOBase and
    provides a read-only interface for consuming multipart body content
    until a boundary is encountered.
    """

    def __init__(self, buffer_reader: MultipartStreamBuffer):
        """
        Initialize body stream reader.

        Args:
            buffer_reader (MultipartStreamBuffer): Reader for buffering content from the stream
        """
        self._buffer_reader = buffer_reader
        self._eof = False
        # Cache content/iterator for efficient reading
        self._read_iterator = None
        self._cached_content = None

    @override
    def read(self, size: int = -1) -> bytes:
        """
        Read up to size bytes from the stream.

        For multipart streams, this respects logical boundaries and may return
        more data than requested to maintain boundary integrity.

        Args:
            size (int): Number of bytes to read. If size is omitted, None, or
                negative read as much as possible. Defaults to -1.

        Returns:
            bytes: Data read from the stream.
        """
        if self._eof and not self._cached_content:
            return b""

        # Initialize the iterator if not already done
        if not self._read_iterator:
            self._read_iterator = self._buffer_reader.read_until_boundary()
            self._iterator_exhausted = False

        # If size is omitted, None, or negative read as much as possible
        if not size or size < 0:
            result = bytearray()
            if self._cached_content:
                result.extend(self._cached_content)
                self._cached_content.clear()

            try:
                for chunk in self._read_iterator:
                    result.extend(chunk)
            except StopIteration:
                pass

            self._eof = True
            return bytes(result)

        # For specific size, accumulate chunks until we have enough
        if not self._cached_content:
            self._cached_content = bytearray()

        # Read more chunks until we have enough data
        while len(self._cached_content) < size and not self._iterator_exhausted:
            try:
                chunk = next(self._read_iterator)
                self._cached_content.extend(chunk)
            except StopIteration:
                self._iterator_exhausted = True
                self._eof = True
                break

        if not self._cached_content:
            return b""

        # Return requested amount or all available
        if len(self._cached_content) <= size:
            result = bytes(self._cached_content)
            self._cached_content.clear()
            return result

        # Return partial content
        result = bytes(self._cached_content[:size])
        # More efficient than slicing assignment
        del self._cached_content[:size]
        return result

    @override
    def readable(self) -> bool:
        """Return True if the stream is readable."""
        return not self._eof or bool(self._cached_content)

    @override
    def close(self) -> None:
        """Close the reader."""
        self._eof = True
        self._cached_content = None

    @override
    def seekable(self) -> bool:
        """Return False as this stream does not support seeking."""
        return False

    @override
    def writable(self) -> bool:
        """Return False as this stream is read-only."""
        return False



"""
Base class for AIS Map Style Datasets

Copyright (c) 2024-2025, NVIDIA CORPORATION. All rights reserved.
"""

from typing import List, Union, Dict
from abc import ABC, abstractmethod
from torch.utils.data import Dataset
from aistore.sdk import AISSource, Object


class AISBaseMapDataset(ABC, Dataset):
    """
    A base class for creating map-style AIS Datasets. Should not be instantiated directly. Subclasses
    should implement :meth:`__getitem__` which fetches a samples given a key from the dataset and can optionally
    override other methods from torch Dataset such as :meth:`__len__` and :meth:`__getitems__`.

    Args:
        ais_source_list (Union[AISSource, List[AISSource]]): Single or list of AISSource objects to load data
        prefix_map (Dict(AISSource, List[str]), optional): Map of AISSource objects to list of prefixes that only allows
        objects with the specified prefixes to be used from each source
    """

    def __init__(
        self,
        ais_source_list: Union[AISSource, List[AISSource]],
        prefix_map: Dict[AISSource, Union[str, List[str]]] = {},
    ) -> None:
        if not ais_source_list:
            raise ValueError(
                f"<{self.__class__.__name__}> ais_source_list must be provided"
            )
        self._ais_source_list = (
            [ais_source_list]
            if isinstance(ais_source_list, AISSource)
            else ais_source_list
        )
        self._prefix_map = prefix_map
        self._obj_list = self._create_objects_list()

    def _create_objects_list(self) -> List[Object]:
        """
        Create a list of all the objects in the given URLs and AIS sources.

        Returns:
            List[Object]: List of all the objects in the given URLs and AIS sources
        """
        samples = []

        for source in self._ais_source_list:
            if source not in self._prefix_map or self._prefix_map[source] is None:
                samples.extend(list(source.list_all_objects_iter(prefix="")))
            else:
                prefixes = (
                    [self._prefix_map[source]]
                    if isinstance(self._prefix_map[source], str)
                    else self._prefix_map[source]
                )
                for prefix in prefixes:
                    samples.extend(list(source.list_all_objects_iter(prefix=prefix)))

        return samples

    def get_obj_list(self) -> List[Object]:
        """
        Getter for internal object data list.

        Returns:
            List[Object]: Object data of the dataset
        """
        return self._obj_list

    @abstractmethod
    def __getitem__(self, index):
        pass

    def __getitems__(self, indices: List[int]):
        return [self.__getitem__(index) for index in indices]



#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

import logging
import re
from pathlib import Path
from typing import Iterator, Optional, Tuple, Type, TypeVar, Union
from urllib.parse import urlparse, parse_qs

import braceexpand
import humanize
import requests
import xxhash

from msgspec import msgpack
from pydantic import BaseModel, TypeAdapter
from urllib3.exceptions import MaxRetryError, ReadTimeoutError

from aistore.sdk.const import (
    HEADER_CONTENT_TYPE,
    MSGPACK_CONTENT_TYPE,
    DEFAULT_LOG_FORMAT,
    XX_HASH_SEED,
    QPARAM_PROVIDER,
)
from aistore.sdk.provider import Provider, provider_aliases

T = TypeVar("T")
MASK = 0xFFFFFFFFFFFFFFFF  # 64-bit mask
# fmt: off
GOLDEN_RATIO = 0x9e3779b97f4a7c15
CONST1 = 0xbf58476d1ce4e5b9
CONST2 = 0x94d049bb133111eb
# fmt: on
ROTATION_BITS = 7

# URL parsing regex components
URL_PROVIDERS = "|".join([p.value for p in Provider] + list(provider_aliases))
MAX_BUCKET_PART_LEN = (
    132  # Accommodates constraint of @uuid(32)#namespace(32)/bucket(64)
)
BUCKET_CHARS = r"[A-Za-z0-9@#._-]"


class HttpError(BaseModel):
    """
    Represents an error returned by the API.
    """

    status_code: int
    message: str = ""
    method: str = ""
    url_path: str = ""
    remote_addr: str = ""
    caller: str = ""
    node: str = ""


def probing_frequency(dur: int) -> float:
    """
    Given a timeout, return an interval to wait between retries

    Args:
        dur: Duration of timeout

    Returns:
        Frequency to probe
    """
    freq = min(dur / 8.0, 1.0)
    freq = max(dur / 64.0, freq)
    return max(freq, 0.1)


def read_file_bytes(filepath: str) -> bytes:
    """
    Given a filepath, read the content as bytes
    Args:
        filepath: Existing local filepath

    Returns: Raw bytes
    """
    with open(filepath, "rb") as reader:
        return reader.read()


def _check_path_exists(path: Path) -> None:
    if not path.exists():
        raise ValueError(f"Path: {path} does not exist")


def validate_file(path: Union[str, Path]) -> None:
    """
    Validate that a file exists and is a file
    Args:
        path (str or Path): Path to validate
    Raises:
        ValueError: If path does not exist or is not a file
    """
    if isinstance(path, str):
        path = Path(path)
    _check_path_exists(path)
    if not path.exists():
        raise ValueError(f"Path: {path} does not exist")
    if not path.is_file():
        raise ValueError(f"Path: {path} is a directory, not a file")


def validate_directory(path: Union[str, Path]) -> None:
    """
    Validate that a directory exists and is a directory
    Args:
        path (str or Path): Path to validate
    Raises:
        ValueError: If path does not exist or is not a directory
    """
    if isinstance(path, str):
        path = Path(path)
    _check_path_exists(path)
    if not path.is_dir():
        raise ValueError(f"Path: {path} is a file, not a directory")


def get_file_size(file: Path) -> str:
    """
    Get the size of a file and return it in human-readable format
    Args:
        file: File to read

    Returns:
        Size of file as human-readable string

    """
    return (
        humanize.naturalsize(file.stat().st_size) if file.stat().st_size else "unknown"
    )


def expand_braces(template: str) -> Iterator[str]:
    """
    Given a string template, apply bash-style brace expansion to return a list of strings
    Args:
        template: Valid brace expansion input, e.g. prefix-{0..10..2}-gap-{11..15}-suffix

    Returns:
        Iterator of brace expansion output

    """
    # pylint: disable = fixme
    # TODO Build custom expansion to validate consistent with cmn/cos/template.go TemplateRange
    return braceexpand.braceexpand(template)


def decode_response(
    res_model: Type[T],
    resp: requests.Response,
) -> T:
    """
    Parse response content from the cluster into a Python class,
     decoding with msgpack depending on content type in header

    Args:
        res_model (Type[T]): Resulting type to which the response should be deserialized
        resp (Response): Response from the AIS cluster

    """
    if resp.headers.get(HEADER_CONTENT_TYPE) == MSGPACK_CONTENT_TYPE:
        return msgpack.decode(resp.content, type=res_model)
    return TypeAdapter(res_model).validate_json(resp.text)


def parse_url(url: str) -> Tuple[str, str, str]:
    """
    Parse AIS URLs for bucket and object names.

    Args:
        url (str): Complete URL of the object (e.g., "ais://bucket1/file.txt")

    Returns:
        Tuple[str, str, str]: Provider, bucket name, and object name
    """
    parsed_url = urlparse(url)
    path = parsed_url.path.lstrip("/")
    return parsed_url.scheme, parsed_url.netloc, path


def extract_and_parse_url(msg: str) -> Optional[Tuple[str, str, bool]]:
    """
    Extract provider, bucket, and whether an object is present from raw string.

    Args:
        msg (str): Any string that may contain an AIS FQN.

    Returns:
        Optional[Tuple[str, str, bool]]: (prov, bck, has_obj) if a FQN is found, otherwise None.
    """
    pattern = rf"({URL_PROVIDERS})://({BUCKET_CHARS}{{1,{MAX_BUCKET_PART_LEN}}})(?:(/)|(?!{BUCKET_CHARS}))"
    match = re.search(pattern, msg)
    if not match:
        return None

    prov = match.group(1)
    bck = match.group(2)
    has_obj = bool(match.group(3))

    return prov, bck, has_obj


def get_logger(name: str, log_format: str = DEFAULT_LOG_FORMAT):
    """
    Create or retrieve a logger with the specified configuration.

    Args:
        name (str): The name of the logger.
        log_format (str, optional): Logging format.

    Returns:
        logging.Logger: Configured logger instance.
    """
    logger = logging.getLogger(name)
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(log_format))
        logger.addHandler(handler)
    logger.propagate = False
    return logger


# Translated from:
# http://xoshiro.di.unimi.it/xoshiro256starstar.c
# Scrambled Linear Pseudorandom Number Generators
# David Blackman, Sebastiano Vigna
# https://arxiv.org/abs/1805.01407
# http://www.pcg-random.org/posts/a-quick-look-at-xoshiro256.html
def xoshiro256_hash(seed: int) -> int:
    """
    Xoshiro256-inspired hash function with 64-bit overflow behavior.
    """
    z = (seed + GOLDEN_RATIO) & MASK
    z = (z ^ (z >> 30)) * CONST1 & MASK
    z = (z ^ (z >> 27)) * CONST2 & MASK
    z = (z ^ (z >> 31)) + GOLDEN_RATIO & MASK
    z = (z ^ (z >> 30)) * CONST1 & MASK
    z = (z ^ (z >> 27)) * CONST2 & MASK
    z = (z ^ (z >> 31)) * 5 & MASK
    rotated = ((z << ROTATION_BITS) | (z >> (64 - ROTATION_BITS))) & MASK
    return (rotated * 9) & MASK


def get_digest(name: str) -> int:
    """
    Get the xxhash digest of a given string.
    """
    return xxhash.xxh64(seed=XX_HASH_SEED, input=name.encode("utf-8")).intdigest()


def convert_to_seconds(time_val: Union[str, int]) -> int:
    """
    Converts a time value (e.g., '5s', '10m', '2h', '3d', or 10) to seconds.
    If no unit is provided (e.g., '10' or 10), seconds are assumed.

    Args:
        time_val (Union[str, int]): The time value to convert.

    Returns:
        int: The equivalent time in seconds.

    Raises:
        ValueError: If the format or unit is invalid.
    """
    multipliers = {"s": 1, "m": 60, "h": 3600, "d": 86400}

    if isinstance(time_val, int):
        return time_val

    if not isinstance(time_val, str) or not time_val.strip():
        raise ValueError("Time value must be a non-empty string or integer.")

    time_val = time_val.strip()

    if time_val.isdigit():
        return int(time_val)

    num, unit = time_val[:-1], time_val[-1]

    if unit not in multipliers:
        raise ValueError(f"Unsupported time unit: '{unit}'. Use 's', 'm', 'h', or 'd'.")

    if not num.isdigit():
        raise ValueError(f"Invalid numeric value in time: '{num}'.")

    return int(num) * multipliers[unit]


def is_read_timeout(exc: requests.ConnectionError) -> bool:
    """
    Check if a given ConnectionError was caused by an underlying ReadTimeoutError
    Args:
        exc: Any requests.ConnectionError.

    Returns: If ReadTimeoutError cause the exception.

    """
    if len(exc.args) < 1:
        return False
    inner_exc = exc.args[0]
    # Expect it to be wrapped in urllib's retry
    if not isinstance(inner_exc, MaxRetryError):
        return False
    # urllib3 ReadTimeoutError != requests ReadTimeout
    return isinstance(inner_exc.reason, ReadTimeoutError)


def get_provider_from_request(
    req: Union[requests.Request, requests.PreparedRequest],
) -> Provider:
    """
    Given either a Request or PreparedRequest, return an AIS bucket provider.
    The request property of a `requests.RequestException` can be either of these types,
        so this can be used to find the bucket provider involved in the initial request.

    Args:
        req (Union[requests.Request, requests.PreparedRequest]): Any request or prepared request.

    Returns:
        Parsed AIS bucket Provider Enum.
    """
    if isinstance(req, requests.Request):
        qparams = req.params
    else:
        parsed_url = urlparse(req.url)
        qparams = (
            {k: v[0] for k, v in parse_qs(parsed_url.query).items()}
            if parsed_url.query and isinstance(parsed_url.query, str)
            else None
        )
    if not qparams:
        raise ValueError("Cannot parse provider from request with no query params")
    return Provider.parse(qparams.get(QPARAM_PROVIDER, ""))



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from typing import Optional

from aistore.sdk.batch.extractor.tar_stream_extractor import TarStreamExtractor
from aistore.sdk.batch.extractor.zip_stream_extractor import ZipStreamExtractor
from aistore.sdk.batch.extractor.archive_stream_extractor import ArchiveStreamExtractor


class ExtractorManager:
    """
    Manager class to handle getting proper batch archive stream extractor
    given an output format. Uses singleton pattern to reuse extractor
    instances.
    """

    # For singleton instantiation
    _instance: Optional["ExtractorManager"] = None
    _initialized: bool = False

    def __new__(cls, *args, **kwargs):
        """
        Get the default singleton instance of ExtractorManager.

        Returns:
            ExtractorManager: The default singleton instance
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        """
        Using supported extractors, initializes format map which maps
        string archive formats to `ArchiveStreamExtractor` implementations.
        """
        # Only initialize once
        if self._initialized:
            return

        supported_extractors = (TarStreamExtractor(), ZipStreamExtractor())

        self._format_map = {}
        for extractor in supported_extractors:
            supported_formats = extractor.get_supported_formats()
            for fmt in supported_formats:
                self._format_map[fmt] = extractor

        self._initialized = True

    def get_extractor(self, output_format: str) -> ArchiveStreamExtractor:
        """
        Returns the matching `ArchiveStreamExtractor` child for a given output format.

        Args:
            output_format (str): Output format to extract

        Returns:
            ArchiveStreamExtractor: Extractor which supports extraction of output format

        Throws:
            ValueError: If output format is not supported by any current extractors
        """
        if (
            output_format is not None
            and self._format_map.get(output_format.lower()) is not None
        ):
            return self._format_map[output_format.lower()]

        raise ValueError(f"Unsupported output format type {output_format}")



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import base64
from typing import Type, Tuple
from urllib.parse import urlparse, urlunparse

import cloudpickle
from aistore.sdk.etl.webserver.base_etl_server import ETLServer
from aistore.sdk.const import UTF_ENCODING
from aistore.sdk.errors import InvalidPipelineError


def serialize_class(cls: Type[ETLServer], encoding: str = UTF_ENCODING) -> str:
    """
    Pickle and base64-encode a user-provided ETLServer subclass for transmission.

    Args:
        cls: A subclass of ETLServer to serialize.
        encoding: The string encoding for the Base64 payload.

    Returns:
        A Base64 string containing the pickled class.

    Raises:
        TypeError: If `cls` is not a subclass of ETLServer.
    """
    if not isinstance(cls, type) or not issubclass(cls, ETLServer):
        raise TypeError(f"{cls!r} is not a subclass of ETLServer")
    pickled = cloudpickle.dumps(cls)
    return base64.b64encode(pickled).decode(encoding)


def compose_etl_direct_put_url(
    direct_put_url: str, host_target: str, obj_path: str
) -> str:
    """
    Compose the final direct PUT URL by combining components from multiple URLs.

    Scenarios:
    1) Pipeline stage: direct_put_url has no path → append object path.
    2) Offline transform: direct_put_url has a path → prepend host_target path
       (e.g. "/v1/etl/_object/<etl-name>/<etl-secret>/") to validate request.

    Args:
        direct_put_url (str): Destination node's direct PUT URL, possibly with path/query.
        host_target (str): Base AIS target URL used for scheme and base path.
        obj_path (str): Path of the object to PUT.
    Returns:
        str: Complete direct PUT URL targeting the correct AIS node.
    """
    direct = urlparse(direct_put_url)
    host = urlparse(host_target)

    if direct.path:
        # Case 2: offline transform → prepend host target's path
        final_path = host.path + direct.path
    else:
        # Case 1: pipeline stage → append object path
        final_path = obj_path

    return urlunparse(
        host._replace(
            netloc=direct.netloc,
            path=final_path,
            query=direct.query,  # keep xid or stats query params
        )
    )


def parse_etl_pipeline(pipeline_header: str) -> Tuple[str, str]:
    """
    Parse ETL pipeline from header value with validation.

    Args:
        pipeline_header: Comma-separated pipeline URLs

    Returns:
        Tuple of (first_url, remaining_pipeline_header)
        where remaining_pipeline_header is comma-joined remaining URLs
        or empty string if no remaining stages

    Raises:
        InvalidPipelineError: If pipeline header is malformed
    """
    if not pipeline_header or not pipeline_header.strip():
        return "", ""

    # Validate basic format - check for empty entries
    pipeline_header = pipeline_header.strip()
    entries = [entry.strip() for entry in pipeline_header.split(",")]

    for entry in entries:
        if not entry:
            raise InvalidPipelineError("Pipeline header contains empty entry")

    # Find the first comma efficiently (after validation)
    comma_index = pipeline_header.find(",")

    if comma_index == -1:
        # No comma found, only one URL (already validated)
        return pipeline_header.strip(), ""

    # Extract first URL and remaining pipeline
    first_url = pipeline_header[:comma_index].strip()
    remaining_pipeline = pipeline_header[comma_index + 1 :].strip()

    return first_url, remaining_pipeline



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

"""
Content iterator providers for streaming object data.

This package provides iterators for fetching object content either sequentially
or in parallel using concurrent range-reads.
"""

from aistore.sdk.obj.content_iterator.base import BaseContentIterProvider
from aistore.sdk.obj.content_iterator.sequential import ContentIterProvider
from aistore.sdk.obj.content_iterator.parallel import ParallelContentIterProvider

__all__ = [
    "BaseContentIterProvider",
    "ContentIterProvider",
    "ParallelContentIterProvider",
]



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

from typing import Optional

from aistore.sdk.const import WHITESPACE_CHARS


class SlidingWindowBuffer:
    """
    High-performance sliding window buffer optimized for streaming data processing.

    The buffer automatically manages memory by sliding the window when it reaches
    capacity, keeping only the most recent data plus a configurable overlap for
    cross-chunk pattern detection.

    Args:
        max_size (int): Maximum logical size before sliding window activation
        boundary_size (int): Size of boundary pattern for overlap calculations

    Notes:
        ```
        buffer: [discarded data][----valid data----][unused space]
                                ^start_pos        ^end_pos
        ```
        This class is not thread-safe. Concurrent access requires external
        synchronization.
    """

    def __init__(self, max_size: int, boundary_size: int):
        """
        Initialize sliding window buffer with size limits and boundary considerations.

        Args:
            max_size (int): Maximum logical size before sliding window activation
            boundary_size (int): Size of boundary pattern for overlap calculations
        """
        self.max_size = max_size
        self.boundary_size = boundary_size
        # Allocate max_size mutable buffer
        self.buffer = bytearray(max_size)

        # Logical start and end positions of buffer
        self.start_pos = 0
        self.end_pos = 0
        self.total_processed = 0

    def append(self, data: bytes) -> None:
        """
        Append data to buffer with automatic sliding window management and compaction.

        Args:
            data (bytes): Data to append to the buffer
        """
        data_len = len(data)

        # Slide window if buffer gets too large
        current_size = self.end_pos - self.start_pos
        if current_size + data_len > self.max_size:
            # Keep overlap for boundary detection spanning chunks
            # 2x boundary_size accounts for CRLF + "--" prefix/suffix + boundary string
            overlap = self.boundary_size * 2

            # Ensure we do not try to remove more data than we have
            bytes_to_remove = max(0, current_size - overlap)

            self.start_pos += bytes_to_remove
            self.total_processed += bytes_to_remove

        # After sliding, ensure we still have valid data
        if self.start_pos >= self.end_pos:
            # Edge case: sliding removed all data
            self.start_pos = self.end_pos = 0

        # Check if we need to compact buffer
        if self.end_pos + data_len > len(self.buffer):
            self._compact_buffer()

            # If still not enough space after compacting, expand buffer
            if self.end_pos + data_len > len(self.buffer):
                new_size = max(len(self.buffer) * 2, self.end_pos + data_len)
                new_buffer = bytearray(new_size)
                new_buffer[: self.end_pos] = memoryview(self.buffer)[: self.end_pos]
                self.buffer = new_buffer

        # Copy data directly into buffer
        self.end_pos += data_len
        self.buffer[self.end_pos - data_len : self.end_pos] = data

    def _compact_buffer(self) -> None:
        """Compact buffer by moving valid data to beginning and updating position tracking."""
        current_data_len = self.end_pos - self.start_pos
        if current_data_len > 0 and self.start_pos > 0:
            # Use memoryview for efficient copying
            source = memoryview(self.buffer)[self.start_pos : self.end_pos]
            self.buffer[:current_data_len] = source

            self.total_processed += self.start_pos
            self.start_pos = 0
            self.end_pos = current_data_len

    def find(self, pattern: bytes, start: int = 0) -> int:
        """
        Find pattern in buffer and return absolute position in stream.

        Args:
            pattern (bytes): Pattern to search for
            start (int): Starting offset for search. Defaults to 0.

        Returns:
            int: Absolute position if found, -1 if not found
        """
        relative_pos = self.find_relative(pattern, start)
        return (
            self.total_processed + self.start_pos + relative_pos
            if relative_pos != -1
            else -1
        )

    def find_relative(self, pattern: bytes, start: int = 0) -> int:
        """
        Find pattern in buffer and return position relative to current buffer window.

        Args:
            pattern (bytes): Pattern to search for
            start (int): Starting offset for search. Defaults to 0.

        Returns:
            int: Relative position if found, -1 if not found
        """
        search_start = self.start_pos + start
        if search_start >= self.end_pos:
            return -1

        # Direct find - position is relative to entire buffer
        pos = self.buffer.find(pattern, search_start, self.end_pos)

        if pos != -1:
            # Convert to relative position
            return pos - self.start_pos
        return -1

    def _get_data(self, start: int = 0, end: Optional[int] = None) -> memoryview:
        """
        Get a data slice without copying using memoryview for efficiency.

        WARNING: The returned memoryview is only valid until the next call to
        append(), consume(), or any other method that modifies the buffer.
        For long-term storage, use get_data_bytes() instead.

        Safe usage pattern:
            ```
            view = self._get_data()
            data = view.tobytes()  # Immediately copy
            # Don't use 'view' after this point
            ```

        Args:
            start (int): Starting position. Defaults to 0.
            end (int, optional): Ending position, None for end of buffer.

        Returns:
            memoryview: Memory view of requested data range (TEMPORARY)
        """
        actual_start = self.start_pos + start
        if end is None:
            actual_end = self.end_pos
        else:
            actual_end = self.start_pos + end

        actual_end = min(actual_end, self.end_pos)
        actual_start = min(actual_start, self.end_pos)

        return memoryview(self.buffer)[actual_start:actual_end]

    def get_data_bytes(self, start: int = 0, end: Optional[int] = None) -> bytes:
        """
        Get data as bytes by creating a copy of the requested range.

        Args:
            start (int): Starting position. Defaults to 0.
            end (int, optional): Ending position, None for end of buffer

        Returns:
            bytes: Copy of requested data range
        """
        return self._get_data(start, end).tobytes()

    def consume(self, length: int) -> bytes:
        """
        Consume and return data from start of buffer, advancing the window.

        Args:
            length (int): Number of bytes to consume

        Returns:
            bytes: Consumed data
        """
        if length <= 0:
            return b""

        actual_length = min(length, self.end_pos - self.start_pos)

        self.start_pos += actual_length
        self.total_processed += actual_length

        return memoryview(self.buffer)[
            self.start_pos - actual_length : self.start_pos
        ].tobytes()

    def skip_leading_whitespace(self) -> int:
        """
        Skip leading whitespace characters and return count of bytes skipped.

        Returns:
            int: Number of whitespace bytes that were skipped and consumed
        """
        buffer_view = self._get_data()
        skip_count = 0

        for byte in buffer_view:
            if byte in WHITESPACE_CHARS:
                skip_count += 1
            else:
                break

        if skip_count > 0:
            self.consume(skip_count)

        return skip_count

    def __len__(self) -> int:
        """
        Return current logical length of valid data in buffer.

        Returns:
            int: Number of bytes of valid data
        """
        return self.end_pos - self.start_pos

    def clear(self) -> None:
        """Clear the buffer and update total processed byte counter."""
        self.total_processed += self.end_pos - self.start_pos
        self.start_pos = 0
        self.end_pos = 0



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import os
from urllib.parse import unquote, quote
from typing import Tuple

import requests
from flask import Flask, request, Response, jsonify

from aistore.sdk.etl.webserver.base_etl_server import ETLServer
from aistore.sdk.etl.webserver.utils import (
    compose_etl_direct_put_url,
    parse_etl_pipeline,
)
from aistore.sdk.errors import InvalidPipelineError
from aistore.sdk.const import (
    HEADER_NODE_URL,
    STATUS_OK,
    QPARAM_ETL_ARGS,
    QPARAM_ETL_FQN,
    HEADER_DIRECT_PUT_LENGTH,
    STATUS_INTERNAL_SERVER_ERROR,
)


class FlaskServer(ETLServer):
    """
    Flask server implementation for ETL transformations.
    Compatible with environments where Flask is preferred over FastAPI.
    """

    def __init__(self, host: str = "0.0.0.0", port: int = 8000):
        super().__init__()
        self.host = host
        self.port = port
        self.app = Flask(__name__)
        self._register_routes()

    def _register_routes(self):
        self.app.add_url_rule("/health", view_func=self._health, methods=["GET"])
        self.app.add_url_rule(
            "/<path:path>", view_func=self._handle_request, methods=["GET", "PUT"]
        )

    def _health(self):
        return Response(response=b"Running", status=200)

    def _handle_request(self, path):
        try:
            transformed = None
            if request.method == "GET":
                transformed = self._handle_get(path)
            else:
                transformed = self._handle_put(path)

            pipeline_header = request.headers.get(HEADER_NODE_URL)
            self.logger.debug("pipeline_header: %r", pipeline_header)
            if pipeline_header:
                first_url, remaining_pipeline = parse_etl_pipeline(pipeline_header)
                if first_url:
                    status_code, transformed, direct_put_length = self._direct_put(
                        first_url, transformed, remaining_pipeline, path
                    )
                    return Response(
                        response=transformed,
                        status=status_code,
                        headers=(
                            {HEADER_DIRECT_PUT_LENGTH: str(direct_put_length)}
                            if direct_put_length != 0
                            else {}
                        ),
                    )

            return Response(
                response=transformed,
                status=STATUS_OK,
                content_type=self.get_mime_type(),
            )

        except InvalidPipelineError as e:
            self.logger.error("Invalid pipeline header: %s", str(e))
            return Response(f"Invalid pipeline header: {str(e)}", status=400)
        except FileNotFoundError as exc:
            fs_path = exc.filename or path
            self.logger.error(
                "Error processing object %r: file not found at %r",
                path,
                fs_path,
            )
            return (
                jsonify(
                    {
                        "error": (
                            f"Error processing object {path!r}: file not found at {fs_path!r}. "
                        )
                    }
                ),
                404,
            )
        except requests.RequestException as e:
            self.logger.error("Request to AIS target failed: %s", str(e))
            return jsonify({"error": str(e)}), 502
        except Exception as e:
            self.logger.exception("Unhandled error")
            return jsonify({"error": str(e)}), 500

    def _handle_get(self, path):
        etl_args = request.args.get(QPARAM_ETL_ARGS, "").strip()
        fqn = request.args.get(QPARAM_ETL_FQN, "").strip()
        if fqn:
            content = self._get_fqn_content(fqn)
        else:
            obj_path = quote(path, safe="@")
            target_url = f"{self.host_target}/{obj_path}"
            self.logger.debug("Forwarding GET to: %s", target_url)
            resp = requests.get(target_url, timeout=None)
            resp.raise_for_status()
            content = resp.content
        return self.transform(content, path, etl_args)

    def _handle_put(self, path):
        etl_args = request.args.get(QPARAM_ETL_ARGS, "").strip()
        fqn = request.args.get(QPARAM_ETL_FQN, "").strip()
        if fqn:
            content = self._get_fqn_content(fqn)
        else:
            content = request.get_data()

        return self.transform(content, path, etl_args)

    def _get_fqn_content(self, path: str) -> bytes:
        decoded_path = unquote(path)
        safe_path = os.path.normpath(os.path.join("/", decoded_path.lstrip("/")))
        self.logger.debug("Reading local file: %s", safe_path)
        with open(safe_path, "rb") as f:
            return f.read()

    def _direct_put(
        self,
        direct_put_url: str,
        data: bytes,
        remaining_pipeline: str = "",
        path: str = "",
    ) -> Tuple[int, bytes, int]:
        """
        Sends the transformed object directly to the specified AIS node (`direct_put_url`),
        eliminating the additional network hop through the original target.
        Used only in bucket-to-bucket offline transforms.

        Args:
            direct_put_url: The first URL in the ETL pipeline
            data: The transformed data to send
            remaining_pipeline: Comma-separated remaining pipeline stages to pass as header
            path: The path of the object.
        Returns:
            status code, transformed data, length of the transformed data (if any)
        """
        try:
            url = compose_etl_direct_put_url(direct_put_url, self.host_target, path)
            headers = {}
            if remaining_pipeline:
                headers[HEADER_NODE_URL] = remaining_pipeline

            resp = self.client_put(url, data, headers=headers)
            return self.handle_direct_put_response(resp, data)

        except Exception as e:
            error = str(e).encode()
            self.logger.error("Exception in direct put to %s: %s", direct_put_url, e)
            return STATUS_INTERNAL_SERVER_ERROR, error, 0

    # Example Gunicorn command to run this server:
    # command: ["gunicorn", "your_module:flask_app", "--bind", "0.0.0.0:8000", "--workers", "4"]
    def start(self):
        self.logger.info("Starting Flask ETL server at %s:%d", self.host, self.port)
        self.app.run(
            host=self.host, port=self.port, threaded=True
        )  # threaded=True for multi-threaded support
